function getOriginalCodeAndFiller(address)
  local original,filler

  if type(address)~='number' then
    address=getAddressSafe(address)
  end

  if address==nil then
    return nil, 'invalid address'
  end

  local sl=createStringList()
  local d=createDisassembler()
  local size=0
  while size<5 do
    d.disassemble(address)
    local ldd=d.LastDisassembleData
    local inst=ldd.opcode..' '..ldd.parameters
    sl.add(inst)
    size=size+#ldd.bytes
    address=address+#ldd.bytes
  end

  original=sl.Text
  if size-5>0 then
    filler=string.format("nop %x", size-5)
  else
    filler=''
  end

  sl.destroy()
  d.destroy()
  return original,filler
end


function hookSpeedFunctions()
  if speedhack and speedhack.processid==getOpenedProcessID() then
    messageDialog('Can not re-enable speedhack on the same process twice', mtError, mbOK)
    return false
  end

  local result, data=autoAssemble([[
    alloc(speedhack_wantedspeed,8)
    registersymbol(speedhack_wantedspeed)
    speedhack_wantedspeed:
    dq (double)1

    alloc(cespinlock,32)
    cespinlock:
    lock bts [rdi],0 //on linux change to rdi
    jc cespinlock_wait
    ret
    cespinlock_wait:
    pause
    jmp cespinlock

{$c}

    typedef struct _cecs
    {
      volatile int locked;
      volatile int threadid;
      volatile int lockcount;
    } cecs, *Pcecs;

    extern void cespinlock(int *lock);
    extern int pthread_self();

    void csenter(cecs *cs)
    {
      if ((cs->locked) && (cs->threadid==pthread_self()))
      {
        cs->lockcount++;
        return;
      }

      cespinlock(&cs->locked);
      cs->threadid=pthread_self();
      cs->lockcount++;
    }

    void csleave(cecs *cs)
    {
      cs->lockcount--;
      if (cs->lockcount==0)
      {
        cs->threadid=0;
        cs->locked=0;
      }

    }

{$asm}

  ]])


  local mataddress=getAddressSafe('libsystem_kernel.dylib.mach_absolute_time')
  if mataddress==nil then
    waitforExports()
    mataddress=getAddressSafe('libsystem_kernel.dylib.mach_absolute_time')

    if (mataddress==nil) then
      reinitializeSymbolhandler()
      if (mataddress==nil) then
        messageDialog('Failure finding libsystem_kernel.dylib.mach_absolute_time', mtError, mbOK)
        return false
      end
    end
  end

  local gtodaddress=getAddressSafe('libsystem_c.dylib.gettimeofday')

  local mat_originalcode,mat_filler=getOriginalCodeAndFiller(mataddress)
  local gtod_originalcode,gtod_filler=getOriginalCodeAndFiller(gtodaddress)
 

--speedhack does not disable. Just sets speed to 1 when done

  local s=string.format([[
alloc(mat_originalcode,64,"libsystem_kernel.dylib")
label(mat_returnhere)
label(mathook_exit)

alloc(gtod_originalcode,64,"libsystem_c.dylib")
label(gtod_returnhere)
label(gtodhook_exit)


{$c}
#include <stddef.h>
#include <celib.h>

struct timeval {
  long tv_sec;
  long tv_usec;
};

uint64_t gtod_originalcode(struct timeval *tp, void *tz);
uint64_t mat_originalcode(void);
double mat_speed=1; //internal helper for new_mach_absolute_time
double gtod_speed=1;
uint64_t mat_initialtime=0;
uint64_t mat_initialoffset=0;


uint64_t mat_lastresult=0;
uint64_t mat_lastresult2=0;


uint64_t gtod_initialtime=0; //just reboot after some years...
uint64_t gtod_initialoffset=0;

int bugged=0;
int speedchangecount=0;

cecs sh_cs; //one cs for all time functions

extern double speedhack_wantedspeed;
extern void csenter(cecs *cs);
extern void csleave(cecs *cs);

/*uint64_t callerlist[4096];
int callerlistsize=0;*/

uint64_t new_mach_absolute_time(void)
{
  uint64_t caller=(uint64_t)__builtin_return_address(1);
  uint64_t caller2=(uint64_t)__builtin_return_address(2);

  
  uint64_t newtime;

  uint64_t currenttime=mat_originalcode();

  double wantedspeed=speedhack_wantedspeed; //small issue with tcc where you can not compare against extern directly

  csenter(&sh_cs);



  
  if ((sh_cs.lockcount==1) && ((caller<0x7FFF00000000ULL) || (caller2<0x7FFF00000000ULL)) ) //not on recursive calls
  {
/*
    int i,found=0;
    for (i=0; i<callerlistsize; i++)
    {
      if (callerlist[i]==caller)
      {
        found=1;
        break;
      }
    }
    if (found==0)
    {
      callerlist[callerlistsize]=caller;
      callerlistsize++;
    }*/



    if (mat_initialtime==0)
    {
      mat_initialtime=currenttime;
      mat_initialoffset=currenttime;
    }


    newtime=(currenttime-mat_initialtime);
    newtime=newtime*mat_speed;
    newtime+=mat_initialoffset;
    if (mat_speed!=wantedspeed)
    {
       //the user wants to change the speed
       mat_initialoffset=newtime;
       mat_initialtime=currenttime;
       mat_speed=wantedspeed;
    }


    mat_lastresult2=currenttime;
    mat_lastresult=newtime;
  
    if (newtime<currenttime)
    {
      bugged=1;
      if (mat_initialoffset<currenttime)
        bugged=2;
    }
  
  }
  else
    newtime=currenttime;

  csleave(&sh_cs);


  return newtime;

}


//gettimeofday

int new_gettimeofday(struct timeval *tp, void *tz)
{
  uint64_t newtime;
  int r;
  double wantedspeed=speedhack_wantedspeed; //small issue with tcc where you can not compare against extern directly


  csenter(&sh_cs);
  r=gtod_originalcode(tp, tz);

  if (r==0)
  {
    uint64_t currenttime=tp->tv_sec*1000000+tp->tv_usec;
    if (gtod_initialtime==0)
    {
      gtod_initialtime=currenttime;
      gtod_initialoffset=currenttime;
    }
    newtime=(currenttime-gtod_initialtime) * gtod_speed;
    newtime+=gtod_initialoffset;

    if (newtime<currenttime)
      newtime=currenttime;

    tp->tv_usec=newtime %% 1000000;
    tp->tv_sec=newtime / 1000000;


    if (gtod_speed!=wantedspeed)
    {
       //the user wants to change the speed
       gtod_initialoffset =newtime;
       gtod_initialtime=currenttime;
       gtod_speed=wantedspeed;
    }



  }
  csleave(&sh_cs);

  return r;
}

{$asm}

//mat
mat_originalcode:
%s

mathook_exit:
jmp mat_returnhere

libsystem_kernel.dylib.mach_absolute_time:
jmp new_mach_absolute_time
%s

mat_returnhere:


//gtod
gtod_originalcode:
%s

gtodhook_exit:
jmp gtod_returnhere

libsystem_c.dylib.gettimeofday:
jmp new_gettimeofday
%s

gtod_returnhere:


]],mat_originalcode, mat_filler, gtod_originalcode, gtod_filler)
 -- print(s)
 -- return false

  local result, data=autoAssemble(s)
  if result then
    speedhack={}
    speedhack.processid=getOpenedProcessID()
  else
    print("error:")
    print(s)
    print("---------------")
    print(data)
    print("---------------")
  end

  return result
end

MainForm.cbSpeedhack.OnChange=function(s)
  if s.Checked then
    --print("enabling speedhack")
    local canEnable=(speedhack and speedhack.processid==getOpenedProcessID()) or hookSpeedFunctions()

    if canEnable then
      MainForm.Panel14.Visible=true
      MainForm.btnSetSpeedhack2.OnClick=function(b)
        if speedhack and speedhack.processid==getOpenedProcessID() then
          writeDouble("speedhack_wantedspeed", tonumber(MainForm.editSH2.Text))
        end
      end
    else
      s.Checked=false
    end
  else
    --print("disabling speedhack")
    if speedhack and speedhack.processid==getOpenedProcessHandle() then
      writeDouble("speedhack_wantedspeed", 1)
    end

    MainForm.Panel14.Visible=false
  end
end

